# -*- coding: utf-8 -*-
"""
group aware role manager, returns roles assigned to group a principal
is a member of, in addition to the explicit roles assigned directly
to the principal.

"""
from AccessControl import ClassSecurityInfo
from AccessControl.requestmethod import postonly
from Acquisition import aq_get
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.class_init import InitializeClass
from App.special_dtml import DTMLFile
from Products.PlonePAS.interfaces.capabilities import IAssignRoleCapability
from Products.PlonePAS.utils import getGroupsForPrincipal
from Products.PluggableAuthService.permissions import ManageUsers
from Products.PluggableAuthService.plugins.ZODBRoleManager \
    import ZODBRoleManager
from zope.interface import implementer


def manage_addGroupAwareRoleManager(self, id, title='', RESPONSE=None):
    """
    this is a doc string
    """
    garm = GroupAwareRoleManager(id, title)
    self._setObject(garm.getId(), garm)

    if RESPONSE is not None:
        RESPONSE.redirect('manage_workspace')

manage_addGroupAwareRoleManagerForm = DTMLFile(
    '../zmi/GroupAwareRoleManagerForm', globals())


@implementer(IAssignRoleCapability)
class GroupAwareRoleManager(ZODBRoleManager):

    meta_type = "Group Aware Role Manager"
    security = ClassSecurityInfo()

    def updateRolesList(self):
        role_holder = aq_parent(aq_inner(self._getPAS()))
        for role in getattr(role_holder, '__ac_roles__', ()):
            if role not in ('Anonymous', 'Authenticated') and \
                    role not in self._roles:
                try:
                    self.addRole(role)
                except KeyError:
                    pass

    # don't blow up if manager already exists; mostly for ZopeVersionControl
    def manage_afterAdd(self, item, container):
        try:
            self.addRole('Manager')
        except KeyError:
            pass

        if item is self:
            self.updateRolesList()

    @security.protected(ManageUsers)
    def assignRoleToPrincipal(self, role_id, principal_id, REQUEST=None):
        try:
            return ZODBRoleManager.assignRoleToPrincipal(
                self,
                role_id,
                principal_id
            )
        except KeyError:
            # Lazily update our roles list and try again
            self.updateRolesList()
            return ZODBRoleManager.assignRoleToPrincipal(
                self,
                role_id,
                principal_id
            )

    @security.protected(ManageUsers)
    def assignRolesToPrincipal(self, roles, principal_id, REQUEST=None):
        """ Assign a specific set of roles, and only those roles, to a
        principal.

        o no return value

        o Raise KeyError if a role_id is unknown.
        """
        for role_id in roles:
            if role_id not in ('Authenticated', 'Anonymous', 'Owner'):
                try:
                    # raise KeyError if unknown!
                    self._roles[role_id]
                except KeyError:
                    # Lazily update our roles list and try again
                    self.updateRolesList()
                    if role_id in self._roles:
                        # check if this role is managed by this plugin, and
                        # set it
                        self._roles[role_id]

        self._principal_roles[principal_id] = tuple(roles)

    assignRolesToPrincipal = postonly(assignRolesToPrincipal)

    @security.private
    def getRolesForPrincipal(self, principal, request=None):
        """ See IRolesPlugin.
        """
        roles = set([])
        principal_ids = set([])
        # Some services need to determine the roles obtained from groups
        # while excluding the directly assigned roles.  In this case
        # '__ignore_direct_roles__' = True should be pushed in the request.
        request = aq_get(self, 'REQUEST', None)
        if request is None \
           or not request.get('__ignore_direct_roles__', False):
            principal_ids.add(principal.getId())

        # Some services may need the real roles of an user but **not**
        # the ones he got through his groups. In this case, the
        # '__ignore_group_roles__'= True should be previously pushed
        # in the request.
        plugins = self._getPAS()['plugins']
        if request is None \
           or not request.get('__ignore_group_roles__', False):
            principal_ids.update(
                getGroupsForPrincipal(principal, plugins, request)
            )
        for pid in principal_ids:
            roles.update(self._principal_roles.get(pid, ()))
        return tuple(roles)

    # implement IAssignRoleCapability

    def allowRoleAssign(self, user_id, role_id):
        """True iff this plugin will allow assigning a certain user a
        certain role.

        Note that at least currently this only checks if the role_id
        exists.  If it exists, this method returns True.  Nothing is
        done with the user_id parameter.  This might be wrong.  See
        http://dev.plone.org/plone/ticket/7762
        """
        present = self.getRoleInfo(role_id)
        if present:
            # if we have a role, we can assign it
            # slightly naive, but should be okay.
            return 1

        return 0

    def listRoleIds(self):
        self.updateRolesList()
        return ZODBRoleManager.listRoleIds(self)

    def listRoleInfo(self):
        self.updateRolesList()
        return ZODBRoleManager.listRoleInfo(self)

    def getRoleInfo(self, role_id):
        if role_id not in self._roles:
            self.updateRolesList()
        return ZODBRoleManager.getRoleInfo(self, role_id)


InitializeClass(GroupAwareRoleManager)



# -*- coding: utf-8 -*-
PROJECTNAME = 'PlonePAS'
GLOBALS = globals()

DEFAULT_CHALLENGE_PROTOCOL = ['http']
DEFAULT_PROTO_MAPPING = {
    'WebDAV': DEFAULT_CHALLENGE_PROTOCOL,
    'FTP': DEFAULT_CHALLENGE_PROTOCOL,
    'XML-RPC': DEFAULT_CHALLENGE_PROTOCOL
}

# Settings for member image resize quality
HAS_PIL = True
try:
    from PIL import Image
    PIL_SCALING_ALGO = Image.ANTIALIAS
except ImportError:
    PIL_SCALING_ALGO = None
    HAS_PIL = False

PIL_QUALITY = 88
MEMBER_IMAGE_SCALE = (75, 100)
IMAGE_SCALE_PARAMS = {
    'scale': MEMBER_IMAGE_SCALE,
    'quality': PIL_QUALITY,
    'algorithm': PIL_SCALING_ALGO,
    'default_format': 'PNG'
}



# -*- coding: utf-8 -*-
from Products.PlonePAS.config import IMAGE_SCALE_PARAMS
from Products.PluggableAuthService.interfaces.plugins import IGroupsPlugin
from cStringIO import StringIO
from urllib import quote as url_quote
from urllib import unquote as url_unquote


def unique(iterable):
    d = {}
    for i in iterable:
        d[i] = None
    return d.keys()


def cleanId(id):
    """'url_quote' turns strange chars into '%xx', which is not a valid char
    for ObjectManager. Here we encode '%' into '-' (and '-' into '--' as
    escaping).
    De-clean is possible; see 'decleanId'.
    Assumes that id can start with non-alpha(numeric), which is true.
    """
    __traceback_info__ = (id,)
    if id:
        # note: we provide the 'safe' param to get '/' encoded
        if isinstance(id, unicode):
            id = id.encode('utf-8')
        return url_quote(id, '').replace('-', '--').replace('%', '-')
    return ''


def decleanId(id):
    """Reverse cleanId."""
    if id:
        id = id.replace('--', '\x00').replace('-', '%').replace('\x00', '-')
        return url_unquote(id)
    return ''


def scale_image(image_file, max_size=None, default_format=None):
    """Scales an image down to at most max_size preserving aspect ratio
    from an input file

        >>> from Products.PlonePAS import config
        >>> import os
        >>> from StringIO import StringIO
        >>> from Products.PlonePAS.utils import scale_image
        >>> from PIL import Image

    Let's make a couple test images and see how it works (all are
    100x100), the gif is palletted mode::

        >>> pas_path = os.path.dirname(config.__file__)
        >>> pjoin = os.path.join
        >>> path = pjoin(pas_path, 'tests', 'images')
        >>> orig_jpg = open(pjoin(path, 'test.jpg'), 'rb')
        >>> orig_png = open(pjoin(path, 'test.png'), 'rb')
        >>> orig_gif = open(pjoin(path, 'test.gif'), 'rb')

    We'll also make some evil non-images, including one which
    masquerades as a jpeg (which would trick OFS.Image)::

        >>> invalid = StringIO('<div>Evil!!!</div>')
        >>> sneaky = StringIO('\377\330<div>Evil!!!</div>')

    OK, let's get to it, first check that our bad images fail:

        >>> scale_image(invalid, (50, 50))
        Traceback (most recent call last):
        ...
        IOError: cannot identify image file...
        >>> scale_image(sneaky, (50, 50))
        Traceback (most recent call last):
        ...
        IOError: cannot identify image file...

    Now that that's out of the way we check on our real images to make
    sure the format and mode are preserved, that they are scaled, and that they
    return the correct mimetype::

        >>> new_jpg, mimetype = scale_image(orig_jpg, (50, 50))
        >>> img = Image.open(new_jpg)
        >>> img.size
        (50, 50)
        >>> img.format
        'JPEG'
        >>> mimetype
        'image/jpeg'

        >>> new_png, mimetype = scale_image(orig_png, (50, 50))
        >>> img = Image.open(new_png)
        >>> img.size
        (50, 50)
        >>> img.format
        'PNG'
        >>> mimetype
        'image/png'

        >>> new_gif, mimetype = scale_image(orig_gif, (50, 50))
        >>> img = Image.open(new_gif)
        >>> img.size
        (50, 50)
        >>> img.format
        'GIF'
        >>> img.mode
        'P'
        >>> mimetype
        'image/gif'

    We should also preserve the aspect ratio by scaling to the given
    width only unless told not to (we need to reset out files before
    trying again though::

        >>> orig_jpg.seek(0)
        >>> new_jpg, mimetype = scale_image(orig_jpg, (70, 100))
        >>> img = Image.open(new_jpg)
        >>> img.size
        (70, 70)

        >>> orig_jpg.seek(0)
        >>> new_jpg, mimetype = scale_image(orig_jpg, (70, 50))
        >>> img = Image.open(new_jpg)
        >>> img.size
        (50, 50)

    """
    from PIL import Image

    if max_size is None:
        max_size = IMAGE_SCALE_PARAMS['scale']
    if default_format is None:
        default_format = IMAGE_SCALE_PARAMS['default_format']
    # Make sure we have ints
    size = (int(max_size[0]), int(max_size[1]))
    # Load up the image, don't try to catch errors, we want to fail miserably
    # on invalid images
    image = Image.open(image_file)
    # When might image.format not be true?
    format = image.format
    mimetype = 'image/%s' % format.lower()

    # from Archetypes ImageField
    # consider image mode when scaling
    # source images can be mode '1','L,','P','RGB(A)'
    # convert to greyscale or RGBA before scaling
    # preserve palletted mode (but not pallette)
    # for palletted-only image formats, e.g. GIF
    # PNG compression is OK for RGBA thumbnails
    original_mode = image.mode
    if original_mode == '1':
        image = image.convert('L')
    elif original_mode == 'P':
        image = image.convert('RGBA')
    # Rescale in place with an method that will not alter the aspect ratio
    # and will only shrink the image not enlarge it.
    image.thumbnail(size, resample=IMAGE_SCALE_PARAMS['algorithm'])
    # preserve palletted mode for GIF and PNG
    if original_mode == 'P' and format in ('GIF', 'PNG'):
        image = image.convert('P')
    # Save
    new_file = StringIO()
    image.save(new_file, format, quality=IMAGE_SCALE_PARAMS['quality'])
    new_file.seek(0)
    # Return the file data and the new mimetype
    return new_file, mimetype


def getGroupsForPrincipal(principal, plugins, request=None):
    groups = set()
    for iid, plugin in plugins.listPlugins(IGroupsPlugin):
        groups.update(plugin.getGroupsForPrincipal(principal, request))
    return list(groups)


def safe_unicode(value, encoding='utf-8'):
    """Converts a value to unicode, even it is already a unicode string.
    """
    if isinstance(value, unicode):
        return value
    elif isinstance(value, basestring):
        try:
            value = unicode(value, encoding)
        except UnicodeDecodeError:
            value = value.decode('utf-8', 'replace')
    return value


# Imported from Products.CMFCore.MemberdataTool as it has now been removed.
class CleanupTemp:
    """Used to cleanup _v_temps at the end of the request."""

    def __init__(self, tool):
        self._tool = tool

    def __del__(self):
        try:
            del self._tool._v_temps
        except (AttributeError, KeyError):
            # The object has already been deactivated.
            pass



# -*- coding: utf-8 -*-
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent


class IUserInitialLoginInEvent(IUserLoggedInEvent):
    """A user logs in for the first time in the portal.
    """



# -*- coding: utf-8 -*-
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.interfaces.browser import IPASMemberView
from plone.memoize.instance import memoize
from zope.interface import implementer
from zope.publisher.browser import BrowserView


@implementer(IPASMemberView)
class PASMemberView(BrowserView):

    @memoize
    def info(self, userid=None):
        pm = getToolByName(self.context, 'portal_membership')
        result = pm.getMemberInfo(memberId=userid)
        if result is None:
            # No such member: removed?  We return something useful anyway.
            return {
                'username': userid,
                'description': '',
                'language': '',
                'home_page': '',
                'name_or_id': userid,
                'location': '',
                'fullname': ''
            }
        result['name_or_id'] = result.get('fullname') or \
            result.get('username') or userid
        return result



# -*- coding: utf-8 -*-
from Products.PluggableAuthService.interfaces import plugins
from Products.PluggableAuthService.interfaces.plugins import Interface


class IUserIntrospection(Interface):
    """
    Introspect users in a user source, api users need to be careful as
    all sources may or not support this interface.

    Realistically this can only be done by authentication sources, or
    plugins which have intimate knowledge of such.
    """

    def getUserIds():
        """
        Return a list of user ids
        """

    def getUserNames():
        """
        Return a list of usernames
        """

    def getUsers():
        """
        Return a list of users
        """


class ILocalRolesPlugin(Interface):
    """
    Plugin for determining a user's local roles and object access
    based on local roles.
    """

    def getRolesInContext(user, object):
        """
        Return the list of roles assigned to the user.

        o Include local roles assigned in context of the passed-in object.

        o Include *both* local roles assigned directly to us *and* those
          assigned to our groups.

        o Ripped off from AccessControl.User.BasicUser, which provides
          no other extension mechanism. :(
        """

    def checkLocalRolesAllowed(user, object, object_roles):
        """
        Check whether the user has access to object based
        on local roles. access is determined by a user's local roles
        including one of the object roles.
        """

    def getAllLocalRolesInContext(object):
        """
        Return active all local roles in a context.

        The roles are returned in a dictionary mapping a principal (a
        user or a group) to the set of roles assigned to it.
        """


class IUserManagement(plugins.IUserAdderPlugin):
    """
    Manage users
    """

    def doChangeUser(user_id, password, **kw):
        """
        Change a user's password (differs from role) roles are set in
        the pas engine api for the same but are set via a role
        manager)
        """

    def doDeleteUser(login):
        """
        Remove a user record from a User Manager, with the given login
        and password

        o Return a Boolean indicating whether a user was removed or
          not
        """


class IMutablePropertiesPlugin(Interface):
    """
    Return a property set for a user. Property set can either an
    object conforming to the IMutable property sheet interface or a
    dictionary (in which case the properties are not persistently
    mutable).
    """

    def getPropertiesForUser(user, request=None):
        """
        User -> IMutablePropertySheet || {}

        o User will implement IPropertiedUser.

        o Plugin may scribble on the user, if needed (but must still
          return a mapping, even if empty).

        o May assign properties based on values in the REQUEST object, if
          present
        """

    def setPropertiesForUser(user, propertysheet):
        """
        Set modified properties on the user persistently.

        Raise a ValueError if the property or property value is invalid
        """

    def deleteUser(user_id):
        """
        Remove properties stored for a user
        """


class ISchemaMutablePropertiesPlugin(Interface):

    def addProperty(property_type, property_name, default=None):
        """
        Add a new property to a property provider.
        """



# -*- coding: utf-8 -*-
"""
Password Validation plugin (IValidationPlugin)
Mutable Property Provider
"""
from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PluggableAuthService.interfaces.plugins import IValidationPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from zope.i18nmessageid import MessageFactory
from zope.interface import implementer

_ = MessageFactory('plone')

manage_addPasswordPolicyForm = PageTemplateFile(
    "../zmi/PasswordPolicyForm",
    globals()
)


def manage_addPasswordPolicyPlugin(self, id, title='',
                                   RESPONSE=None, schema=None, **kw):
    """
    Create an instance of a password validation plugin.
    """
    o = PasswordPolicyPlugin(id, title)
    self._setObject(o.getId(), o)

    if RESPONSE is not None:
        return RESPONSE.redirect(
            "%s/manage_workspace?"
            "manage_tabs_message=DefaultPasswordPlugin+plugin+added" %
            self.absolute_url()
        )


@implementer(IValidationPlugin)
class PasswordPolicyPlugin(BasePlugin):
    """Simple Password Policy to ensure password is 5 chars long.
    """

    meta_type = 'Default Plone Password Policy'

    security = ClassSecurityInfo()

    def __init__(self, id, title=''):
        """Create a default plone password policy to ensure 5 char passwords
        """
        self.id = id
        self.title = title

    @security.private
    def validateUserInfo(self, user, set_id, set_info):
        """ See IValidationPlugin. Used to validate password property
        """

        if not set_info:
            return []
        password = set_info.get('password', None)
        if password is None:
            return []
        elif password == '':
            return [{'id': 'password', 'error': _(u'Minimum 5 characters.')}]
        elif len(password) < 5:
            return [
                {
                    'id': 'password',
                    'error': _(
                        u'Your password must contain at least 5 characters.'
                    )
                }
            ]
        else:
            return []

InitializeClass(PasswordPolicyPlugin)



# -*- coding: utf-8 -*-
from Products.PluggableAuthService.interfaces.propertysheets \
    import IPropertySheet


class IMutablePropertySheet(IPropertySheet):

    def canWriteProperty(object, id):
        """ Check if a property can be modified.
        """

    def setProperty(object, id, value):
        """
        """

    def setProperties(object, mapping):
        """
        """


class ISchemaMutablePropertySheet(IMutablePropertySheet):
    pass



# -*- coding: utf-8 -*-
from Products.PlonePAS.tests import base
from Products.PluggableAuthService.PluggableAuthService import \
    _SWALLOWABLE_PLUGIN_EXCEPTIONS
from Products.PluggableAuthService.interfaces.authservice import \
    IPluggableAuthService
from Products.PluggableAuthService.interfaces.events import \
    IPrincipalDeletedEvent
from Products.PluggableAuthService.interfaces.plugins import IRolesPlugin
from zope.component import adapter
from zope.component import getGlobalSiteManager


class BasicOpsTestCase(base.TestCase):

    def afterSetUp(self):
        self.loginAsPortalOwner()
        self.acl_users = self.portal.acl_users

    def compareRoles(self, target, user, roles):
        """
        compareRoles(self, target, user, roles) => do not raise if
        user has exactly the specified roles.

        If target is None, test user roles (no local roles)
        """
        u = self.acl_users.getUser(user)
        if not u:
            raise RuntimeError("compareRoles: Invalid user: '%s'" % user)
        non_roles = ('Authenticated', 'Anonymous', '')
        if target is None:
            user_roles = list(u.getRoles())
        else:
            user_roles = list(u.getRolesInContext(target))
        actual_roles = filter(lambda x: x not in non_roles, user_roles)
        actual_roles.sort()
        wished_roles = list(roles)
        wished_roles.sort()
        if actual_roles == wished_roles:
            return 1
        raise RuntimeError("User %s: Whished roles: %s BUT current "
                           "roles: %s" % (user, wished_roles, actual_roles))

    def createUser(self, login="created_user", password="secret",
                   roles=[], groups=[], domains=()):
        self.acl_users.userFolderAddUser(
            login, password, roles=roles, groups=groups, domains=domains,)

    def test_installed(self):
        self.assertTrue(IPluggableAuthService.providedBy(self.acl_users))

    def test_add(self):
        self.createUser()
        self.assertTrue(self.acl_users.getUser("created_user"))

    def test_edit(self):
        # this will fail unless the PAS role plugin is told it manages
        # the new role.
        self.createUser()
        self.compareRoles(None, "created_user", [])
        self.acl_users.userFolderEditUser(
            "created_user",  # name
            "secret2",  # password
            roles=["Member"],
            groups=["g1"],
            domains=(),)
        self.compareRoles(None, "created_user", ['Member'])

    def test_edit_userDefinedRole(self):
        roleplugins = self.acl_users.plugins.listPlugins(IRolesPlugin)
        for id, plugin in roleplugins:
            try:
                plugin.addRole('r1')
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                pass
            else:
                break

        self.createUser()
        self.compareRoles(None, "created_user", [])
        self.acl_users.userFolderEditUser(
            "created_user",  # name
            "secret2",  # password
            roles=["r1"],
            groups=["g1"],
            domains=(),)
        self.compareRoles(None, "created_user", ['r1'])

    def test_del(self):
        self.createUser()
        self.assertTrue(self.acl_users.getUser("created_user"))
        self.acl_users.userFolderDelUsers(['created_user'])
        self.assertFalse(self.acl_users.getUser("created_user"))

    def test_principal_del_event(self):
        eventsFired = []

        @adapter(IPrincipalDeletedEvent)
        def gotDeletion(event):
            eventsFired.append(event)

        gsm = getGlobalSiteManager()
        gsm.registerHandler(gotDeletion)
        self.createUser()
        self.acl_users.userFolderDelUsers(['created_user'])
        self.assertEqual(len(eventsFired), 1)
        self.assertEqual(eventsFired[0].principal, 'created_user')
        gsm.unregisterHandler(gotDeletion)

    def test_search(self):
        self.createUser("created_user1")
        self.createUser("created_user2")
        mt = self.portal.portal_membership
        retlist = mt.searchForMembers(REQUEST=None, login="created_user1")
        usernames = [user.getUserName() for user in retlist]
        self.assertEqual(len(usernames), 1)
        self.assertTrue("created_user1" in usernames,
                        "'created_user1' not in %s" % usernames)

    def test_setpw(self):
        # there is more than one place where one can set the password.
        # insane. anyway we have to check the patch in pas.py userSetPassword
        # here its checked in the general setup using ZODBUserManager.
        self.createUser()
        uf = self.acl_users
        new_secret = 'new_secret'
        uf.userSetPassword('created_user', new_secret)

        # possible to authenticate with new password?
        from Products.PluggableAuthService.interfaces.plugins \
            import IAuthenticationPlugin
        authenticators = uf.plugins.listPlugins(IAuthenticationPlugin)
        credentials = {'login': 'created_user', 'password': new_secret}
        result = None
        for aid, authenticator in authenticators:
            result = authenticator.authenticateCredentials(credentials)
            if result is not None:
                break
        self.assertTrue(result)

    def test_setProperties(self):
        self.createUser()
        user = self.acl_users.getUser('created_user')
        user.setProperties(fullname='Test User', email='test@example.org')
        self.assertEqual(user.getProperty('fullname'), 'Test User')
        self.assertEqual(user.getProperty('email'), 'test@example.org')

        user.setProperties(
            properties={'fullname': 'Test User2', 'email': 'test2@example.org'}
        )
        self.assertEqual(user.getProperty('fullname'), 'Test User2')
        self.assertEqual(user.getProperty('email'), 'test2@example.org')



# -*- coding: utf-8 -*-
""" Class: CookieCrumblingPlugin

Acts as auth plugin, but injects cookie form credentials as HTTPBasicAuth.
This allows form logins to fall through to parent user folders.

"""
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import aq_base
from App.class_init import InitializeClass
from App.special_dtml import DTMLFile
from OFS.Folder import Folder
from Products.CMFCore.CookieCrumbler import manage_addCC
from Products.PluggableAuthService.interfaces.plugins import IExtractionPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from zope.interface import implementer
import logging

logger = logging.getLogger('PlonePAS')

CC_ID = 'cookie_auth'


def manage_addCookieCrumblingPlugin(self, id, title='',
                                    RESPONSE=None, **kw):
    """
    Create an instance of a cookie crumbling plugin.
    """
    self = self.this()

    o = CookieCrumblingPlugin(id, title, **kw)
    self._setObject(o.getId(), o)
    o = getattr(aq_base(self), id)

    manage_addCC(o, CC_ID)

    if RESPONSE is not None:
        RESPONSE.redirect('manage_workspace')

manage_addCookieCrumblingPluginForm = \
    DTMLFile("../zmi/CookieCrumblingPluginForm", globals())


@implementer(IExtractionPlugin)
class CookieCrumblingPlugin(Folder, BasePlugin):
    """Multi-plugin for injecting HTTP Basic Authentication
    credentials from form credentials.
    """
    meta_type = 'Cookie Crumbling Plugin'

    security = ClassSecurityInfo()

    def __init__(self, id, title=None):
        self._setId(id)
        self.title = title

    def _getCC(self):
        return getattr(aq_base(self), CC_ID, None)

    @security.private
    def extractCredentials(self, request):
        """ Extract basic auth credentials from 'request'.
        """

        try:
            self._getCC().modifyRequest(request, request.RESPONSE)

        except Exception, e:
            logger.error("PlonePAS error: %s", e, exc_info=1)

        return {}

InitializeClass(CookieCrumblingPlugin)



# -*- coding: utf-8 -*-
from AccessControl.Permissions import add_user_folders
from Products.CMFCore.utils import ToolInit
from Products.PlonePAS.pas import patch_pas
from Products.PlonePAS.plugins import autogroup
from Products.PlonePAS.plugins import cookie_handler
from Products.PlonePAS.plugins import crumbler
from Products.PlonePAS.plugins import group
from Products.PlonePAS.plugins import local_role
from Products.PlonePAS.plugins import passwordpolicy
from Products.PlonePAS.plugins import property
from Products.PlonePAS.plugins import role
from Products.PlonePAS.plugins import ufactory
from Products.PlonePAS.plugins import user
from Products.PlonePAS.tools.groupdata import GroupDataTool
from Products.PlonePAS.tools.groups import GroupsTool
from Products.PlonePAS.tools.memberdata import MemberDataTool
from Products.PlonePAS.tools.membership import MembershipTool
from Products.PluggableAuthService import registerMultiPlugin

####################################
# monkey patch pas, the evil happens
patch_pas()

#################################
# new groups tool

#################################
# register plugins with pas
try:
    registerMultiPlugin(user.UserManager.meta_type)
    registerMultiPlugin(group.GroupManager.meta_type)
    registerMultiPlugin(role.GroupAwareRoleManager.meta_type)
    registerMultiPlugin(local_role.LocalRolesManager.meta_type)
    registerMultiPlugin(ufactory.PloneUserFactory.meta_type)
    registerMultiPlugin(property.ZODBMutablePropertyProvider.meta_type)
    registerMultiPlugin(crumbler.CookieCrumblingPlugin.meta_type)
    registerMultiPlugin(cookie_handler.ExtendedCookieAuthHelper.meta_type)
    registerMultiPlugin(autogroup.AutoGroup.meta_type)
    registerMultiPlugin(passwordpolicy.PasswordPolicyPlugin.meta_type)
except RuntimeError:
    # make refresh users happy
    pass


def initialize(context):

    tools = (GroupsTool, GroupDataTool, MembershipTool, MemberDataTool)

    ToolInit(
        'PlonePAS Tool',
        tools=tools,
        icon='tool.gif',
    ).initialize(context)

    context.registerClass(
        role.GroupAwareRoleManager,
        permission=add_user_folders,
        constructors=(
            role.manage_addGroupAwareRoleManagerForm,
            role.manage_addGroupAwareRoleManager),
        visibility=None
    )

    context.registerClass(
        user.UserManager,
        permission=add_user_folders,
        constructors=(
            user.manage_addUserManagerForm,
            user.manage_addUserManager),
        visibility=None
    )

    context.registerClass(
        group.GroupManager,
        permission=add_user_folders,
        constructors=(
            group.manage_addGroupManagerForm,
            group.manage_addGroupManager
        ),
        visibility=None
    )

    context.registerClass(
        ufactory.PloneUserFactory,
        permission=add_user_folders,
        constructors=(
            ufactory.manage_addPloneUserFactoryForm,
            ufactory.manage_addPloneUserFactory),
        visibility=None
    )

    context.registerClass(
        local_role.LocalRolesManager,
        permission=add_user_folders,
        constructors=(
            local_role.manage_addLocalRolesManagerForm,
            local_role.manage_addLocalRolesManager),
        visibility=None
    )

    context.registerClass(
        property.ZODBMutablePropertyProvider,
        permission=add_user_folders,
        constructors=(
            property.manage_addZODBMutablePropertyProviderForm,
            property.manage_addZODBMutablePropertyProvider),
        visibility=None
    )

    context.registerClass(
        crumbler.CookieCrumblingPlugin,
        permission=add_user_folders,
        constructors=(
            crumbler.manage_addCookieCrumblingPluginForm,
            crumbler.manage_addCookieCrumblingPlugin),
        visibility=None
    )

    context.registerClass(
        cookie_handler.ExtendedCookieAuthHelper,
        permission=add_user_folders,
        constructors=(
            cookie_handler.manage_addExtendedCookieAuthHelperForm,
            cookie_handler.manage_addExtendedCookieAuthHelper),
        visibility=None
    )

    context.registerClass(
        autogroup.AutoGroup,
        permission=add_user_folders,
        constructors=(
            autogroup.manage_addAutoGroupForm,
            autogroup.manage_addAutoGroup),
        visibility=None
    )

    context.registerClass(
        passwordpolicy.PasswordPolicyPlugin,
        permission=add_user_folders,
        constructors=(
            passwordpolicy.manage_addPasswordPolicyForm,
            passwordpolicy.manage_addPasswordPolicyPlugin
        ),
        visibility=None
    )



# -*- coding: utf-8 -*-
from App.class_init import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PlonePAS.interfaces.group import IGroupIntrospection
from Products.PluggableAuthService.PropertiedUser import PropertiedUser
from Products.PluggableAuthService.interfaces.plugins import \
    IGroupEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins import IGroupsPlugin
from Products.PluggableAuthService.interfaces.plugins import IPropertiesPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from zope.interface import implementer

manage_addAutoGroupForm = PageTemplateFile("../zmi/AutoGroupForm", globals())


def manage_addAutoGroup(self, id, title='', group='', description='',
                        RESPONSE=None):
    """Add an Auto Group plugin."""

    plugin = AutoGroup(id, title, group, description)
    self._setObject(id, plugin)

    if RESPONSE is not None:
        return RESPONSE.redirect(
            "%s/manage_workspace?manage_tabs_message=AutoGroup+plugin+added"
            % self.absolute_url())


class VirtualGroup(PropertiedUser):
    def __init__(self, id, title='', description=''):
        super(VirtualGroup, self).__init__(id)
        self.id = id
        self.title = title
        self.description = description

    def getId(self):
        return self.id

    def getUserName(self):
        return self.id

    def getName(self):
        return self.id

    def getMemberIds(self, transitive=1):
        return []

    def getRolesInContext(self, context):
        return []

    def getRoles(self):
        return []

    def allowed(self, object, object_roles=None):
        return 0

    def getDomains(self):
        return []

    def isGroup(self):
        return True


@implementer(
    IGroupEnumerationPlugin,
    IGroupsPlugin,
    IGroupIntrospection,
    IPropertiesPlugin
)
class AutoGroup(BasePlugin):
    meta_type = "Automatic Group Plugin"

    _properties = (
        {'id': 'title',
         'label': 'Title',
         'type': 'string',
         'mode': 'w'},
        {'id': 'group',
         'label': 'Group',
         'type': 'string',
         'mode': 'w'},
        {'id': 'description',
         'label': 'Description',
         'type': 'string',
         'mode': 'w'},
    )

    def __init__(self, id, title='', group=None, description=''):
        self._setId(id)
        self.title = title
        self.group = group
        self.description = description

    # IGroupEnumerationPlugin implementation
    def enumerateGroups(self, id=None, exact_match=False, sort_by=None,
                        max_results=None, **kw):
        if kw:
            return []

        if id:
            id = id.lower()
            mygroup = self.group.lower()

            if exact_match and id != mygroup:
                return []

            if not exact_match and id not in mygroup:
                return []

        return [{'id': self.group,
                 'groupid': self.group,
                 'title': self.title,
                 'pluginid': self.getId()}]

    # IGroupsPlugin implementation
    def getGroupsForPrincipal(self, principal, request=None):
        if principal.getUserName() == self.group:
            return ()

        return (self.group,)

    # IGroupIntrospection implementation
    def getGroupById(self, group_id):
        if group_id != self.group:
            return None

        return VirtualGroup(self.group, title=self.title,
                            description=self.description)

    def getGroups(self):
        return [self.getGroupById(id) for id in self.getGroupIds()]

    def getGroupIds(self):
        return [self.group]

    def getGroupMembers(self, group_id):
        return ()

    # IPropertiesPlugin:
    def getPropertiesForUser(self, user, request=None):
        if user == self.group:
            return {'title': self.title,
                    'description': self.description}
        else:
            return {}


InitializeClass(AutoGroup)



# -*- coding: utf-8 -*-
# pas alterations and monkies
from AccessControl import Unauthorized
from AccessControl import getSecurityManager
from AccessControl.PermissionRole import PermissionRole
from AccessControl.Permissions import change_permissions
from AccessControl.Permissions import manage_properties
from AccessControl.Permissions import manage_users as ManageUsers
from AccessControl.requestmethod import postonly
from OFS.Folder import Folder
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.utils import registerToolInterface
from Products.PlonePAS.interfaces.group import IGroupIntrospection
from Products.PlonePAS.interfaces.group import IGroupManagement
from Products.PlonePAS.interfaces.plugins import ILocalRolesPlugin
from Products.PlonePAS.interfaces.plugins import IUserIntrospection
from Products.PlonePAS.interfaces.plugins import IUserManagement
from Products.PlonePAS.patch import ORIG_NAME
from Products.PlonePAS.patch import wrap_method
from Products.PluggableAuthService.PluggableAuthService import \
    PluggableAuthService
from Products.PluggableAuthService.PluggableAuthService import \
    _SWALLOWABLE_PLUGIN_EXCEPTIONS
from Products.PluggableAuthService.events import PrincipalDeleted
from Products.PluggableAuthService.interfaces.authservice import \
    IPluggableAuthService
from Products.PluggableAuthService.interfaces.plugins import \
    IAuthenticationPlugin
from Products.PluggableAuthService.interfaces.plugins import \
    IGroupEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins import \
    IRoleAssignerPlugin
from Products.PluggableAuthService.interfaces.plugins import \
    IUserEnumerationPlugin
from zope.event import notify
import logging

logger = logging.getLogger('PlonePAS')

registerToolInterface('acl_users', IPluggableAuthService)


#################################
# helper functions

def _userSetGroups(pas, user_id, groupnames):
    """method was used at GRUF level, but is used inside this monkies at several
    places too.

    We no longer provide it on PAS to clean up patches

    """
    plugins = pas.plugins
    gtool = getToolByName(pas, "portal_groups")

    member = pas.getUserById(user_id)
    groupnameset = set(groupnames)

    # remove absent groups
    groups = set(gtool.getGroupsForPrincipal(member))
    rmgroups = groups - groupnameset
    for gid in rmgroups:
        try:
            gtool.removePrincipalFromGroup(user_id, gid)
        except KeyError:
            # We could hit a group which does not allow user removal, such as
            # created by our AutoGroup plugin.
            pass

    # add groups
    try:
        groupmanagers = plugins.listPlugins(IGroupManagement)
    except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
        logger.info(
            'PluggableAuthService: Plugin listing error',
            exc_info=1
        )
        groupmanagers = ()

    for group in groupnames:
        for gm_id, gm in groupmanagers:
            try:
                if gm.addPrincipalToGroup(user_id, group):
                    break
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                logger.info(
                    'PluggableAuthService: GroupManagement %s error',
                    gm_id,
                    exc_info=1
                )

#################################
# pas folder monkies - standard zope user folder api or GRUF


def _doAddUser(self, login, password, roles, domains, groups=None, **kw):
    """Masking of PAS._doAddUser to add groups param."""
    _old_doAddUser = getattr(self, getattr(_doAddUser, ORIG_NAME))
    retval = _old_doAddUser(login, password, roles, domains)
    if groups is not None:
        _userSetGroups(self, login, groups)
    return retval


def _doDelUsers(self, names, REQUEST=None):
    """
    Delete users given by a list of user ids.
    Has no return value, like the original (GRUF).
    """
    for name in names:
        self._doDelUser(name)


def _doDelUser(self, id):
    """
    Given a user id, hand off to a deleter plugin if available.
    """
    plugins = self._getOb('plugins')
    userdeleters = plugins.listPlugins(IUserManagement)

    if not userdeleters:
        raise NotImplementedError(
            "There is no plugin that can delete users."
        )

    for userdeleter_id, userdeleter in userdeleters:
        try:
            userdeleter.doDeleteUser(id)
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            pass
        else:
            notify(PrincipalDeleted(id))


def _doChangeUser(self, principal_id, password, roles, domains=(), groups=None,
                  REQUEST=None, **kw):
    """
    Given a principal id, change its password, roles, domains, if
    respective plugins for such exist.

    XXX domains are currently ignored.
    """
    # Might be called with 'None' as password from the Plone UI, in
    # prefs_users_overview when resetPassword is not set.
    if password is not None:
        self.userSetPassword(principal_id, password)

    plugins = self._getOb('plugins')
    rmanagers = plugins.listPlugins(IRoleAssignerPlugin)

    if not (rmanagers):
        raise NotImplementedError("There is no plugin that can modify roles")

    for rid, rmanager in rmanagers:
        rmanager.assignRolesToPrincipal(roles, principal_id)

    if groups is not None:
        _userSetGroups(self, principal_id, groups)

    return True


def userFolderAddUser(self, login, password, roles, domains,
                      groups=None, REQUEST=None, **kw):
    self._doAddUser(login, password, roles, domains, **kw)
    if groups is not None:
        _userSetGroups(self, login, groups)


def _doAddGroup(self, id, roles, groups=None, **kw):
    gtool = getToolByName(self, 'portal_groups')
    return gtool.addGroup(id, roles, groups, **kw)


# for prefs_group_manage compatibility. really should be using tool.
def _doDelGroups(self, names, REQUEST=None):
    gtool = getToolByName(self, 'portal_groups')
    for group_id in names:
        gtool.removeGroup(group_id)


def _doChangeGroup(self, principal_id, roles, groups=None, REQUEST=None, **kw):
    """
    Given a group's id, change its roles, domains, if respective
    plugins for such exist. Domains are currently ignored.

    See also _doChangeUser
    """
    gtool = getToolByName(self, 'portal_groups')
    gtool.editGroup(principal_id, roles, groups, **kw)
    return True


def _updateGroup(self, principal_id, roles=None, groups=None, **kw):
    """
    Given a group's id, change its roles, groups, if respective
    plugins for such exist. Domains are ignored.

    This is not an alias to _doChangeGroup because its params are different
    (slightly).
    """
    return self._doChangeGroup(principal_id, roles, groups, **kw)


def getGroups(self):
    gtool = getToolByName(self, 'portal_groups')
    return gtool.listGroups()


def getGroupNames(self):
    gtool = getToolByName(self, 'portal_groups')
    return gtool.getGroupIds()


def getGroupIds(self):
    gtool = getToolByName(self, 'portal_groups')
    return gtool.getGroupIds()


def getGroup(self, group_id):
    """Like getGroupById in groups tool, but doesn't wrap.
    """
    group = None
    introspectors = self.plugins.listPlugins(IGroupIntrospection)

    if not introspectors:
        raise ValueError('No plugins allow for group management')
    for iid, introspector in introspectors:
        group = introspector.getGroupById(group_id)
        if group is not None:
            break
    return group


def getGroupByName(self, name, default=None):
    ret = self.getGroup(name)
    if ret is None:
        return default
    return ret


def getGroupById(self, id, default=None):
    gtool = getToolByName(self, "portal_groups")
    ret = gtool.getGroupById(id)
    if ret is None:
        return default
    else:
        return ret


def getLocalRolesForDisplay(self, object):
    """This is used for plone's local roles display

    This method returns a tuple (massagedUsername, roles, userType,
    actualUserName).  This method is protected by the 'access content
    information' permission. We may change that if it's too
    permissive...

    A GRUF method originally.
    """
    # Perform security check on destination object
    if not getSecurityManager().checkPermission(manage_properties, object):
        raise Unauthorized(name="getLocalRolesForDisplay")

    return self._getLocalRolesForDisplay(object)


def _getLocalRolesForDisplay(self, object):
    result = []
    # we don't have a PAS-side way to get this
    local_roles = object.get_local_roles()
    for one_user in local_roles:
        username = userid = one_user[0]
        roles = one_user[1]
        userType = 'user'
        if self.getGroup(userid):
            userType = 'group'
        else:
            user = self.getUserById(userid) or self.getUser(username)
            if user:
                username = user.getUserName()
                userid = user.getId()
        result.append((username, roles, userType, userid))
    return tuple(result)


def getUsers(self):
    """
    Return a list of all users from plugins that implement the user
    introspection interface.

    Could potentially be very long.
    """
    # We should have a method that's cheap about returning number of users.
    retval = []
    try:
        introspectors = self.plugins.listPlugins(IUserIntrospection)
    except KeyError:
        return retval

    for iid, introspector in introspectors:
        retval += introspector.getUsers()

    return retval


def canListAllUsers(self):
    plugins = self._getOb('plugins')
    # Do we have multiple user plugins?
    num_enumeration_plugins = plugins.listPlugins(IUserEnumerationPlugin)
    num_introspection_plugins = plugins.listPlugins(IUserEnumerationPlugin)
    return num_enumeration_plugins == num_introspection_plugins


def canListAllGroups(self):
    plugins = self._getOb('plugins')
    # Do we have multiple group plugins?
    num_enumeration_plugins = plugins.listPlugins(IGroupEnumerationPlugin)
    num_introspection_plugins = plugins.listPlugins(IGroupEnumerationPlugin)
    return num_enumeration_plugins == num_introspection_plugins


def userSetPassword(self, userid, password):
    """Emulate GRUF 3 call for password set, for use with PwRT."""
    # used by _doChangeUser
    plugins = self._getOb('plugins')
    managers = plugins.listPlugins(IUserManagement)

    if not managers:
        raise NotImplementedError("There is no plugin that can modify users")

    modified = False
    for mid, manager in managers:
        try:
            manager.doChangeUser(userid, password)
        except RuntimeError:
            # XXX: why silent ignore this Error?
            pass
        else:
            modified = True

    if not modified:
        raise RuntimeError("No user management plugins were able "
                           "to successfully modify the user")


def credentialsChanged(self, user, name, new_password):
    """Notifies the authentication mechanism that this user has changed
    passwords.  This can be used to update the authentication cookie.
    Note that this call should *not* cause any change at all to user
    databases.

    For use by CMFCore.MembershipTool.credentialsChanged
    """
    request = self.REQUEST
    response = request.RESPONSE
    login = name

    self.updateCredentials(request, response, login, new_password)


# for ZopeVersionControl, we need to check 'plugins' for more than
# existence, since it replaces objects (like 'plugins') with SimpleItems
# and calls _delOb, which tries to use special methods of 'plugins'
def _delOb(self, id):
    #
    #   Override ObjectManager's version to clean up any plugin
    #   registrations for the deleted object
    #
    # XXX imo this is a evil one
    #
    plugins = self._getOb('plugins', None)

    if getattr(plugins, 'removePluginById', None) is not None:
        plugins.removePluginById(id)

    Folder._delOb(self, id)


def addRole(self, role):
    plugins = self._getOb('plugins')
    roles = plugins.listPlugins(IRoleAssignerPlugin)

    for plugin_id, plugin in roles:
        try:
            plugin.addRole(role)
            break
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            pass


def getAllLocalRoles(self, context):
    # Perform security check on destination object
    if not getSecurityManager().checkPermission(change_permissions, context):
        raise Unauthorized(name="getAllLocalRoles")
    return self._getAllLocalRoles(context)


def _getAllLocalRoles(self, context):
    plugins = self._getOb('plugins')
    lrmanagers = plugins.listPlugins(ILocalRolesPlugin)

    roles = {}
    for lrid, lrmanager in lrmanagers:
        newroles = lrmanager.getAllLocalRolesInContext(context)
        for k, v in newroles.items():
            if k not in roles:
                roles[k] = set()
            roles[k].update(v)

    return roles


def authenticate(self, name, password, request):
    """See AccessControl.User.BasicUserFolder.authenticate

    Products.PluggableAuthService.PluggableAuthService does not provide this
    method, BasicUserFolder documents it as "Private UserFolder object
    interface". GRUF does provide the method, so not marked as private.

    should be deprecated in future!
    """

    plugins = self.plugins

    try:
        authenticators = plugins.listPlugins(IAuthenticationPlugin)
    except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
        logger.info('PluggableAuthService: Plugin listing error', exc_info=1)
        authenticators = ()

    credentials = {'login': name,
                   'password': password}

    user_id = None

    for authenticator_id, auth in authenticators:
        try:
            uid_and_name = auth.authenticateCredentials(credentials)
            if uid_and_name is not None:
                user_id, name = uid_and_name
                break
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.info(
                'PluggableAuthService: AuthenticationPlugin %s error',
                authenticator_id,
                exc_info=1
            )
            continue

    if not user_id:
        return

    return self._findUser(plugins, user_id, name, request)


def getUserIds(self):
    """method was used at GRUF and is here for bbb. Not good for many users!
    DEPRECATED
    """
    plugins = self.plugins

    try:
        introspectors = plugins.listPlugins(IUserIntrospection)
    except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
        logger.info('PluggableAuthService: Plugin listing error', exc_info=1)
        introspectors = ()

    results = []
    for introspector_id, introspector in introspectors:
        try:
            results.extend(introspector.getUserIds())
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.info(
                'PluggableAuthService: UserIntrospection %s error',
                introspector_id,
                exc_info=1
            )

    return results


def getUserNames(self):
    """method was used at GRUF and is here for bbb. Not good for many users!
    DEPRECATED
    """
    plugins = self.plugins

    try:
        introspectors = plugins.listPlugins(IUserIntrospection)
    except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
        logger.info('PluggableAuthService: Plugin listing error', exc_info=1)
        introspectors = ()

    results = []
    for introspector_id, introspector in introspectors:
        try:
            results.extend(introspector.getUserNames())
        except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
            logger.info(
                'PluggableAuthService: UserIntroSpection plugin %s error',
                introspector_id, exc_info=1)

    return results


def patch_pas():
    # sort alphabetically by patched/added method name
    wrap_method(
        PluggableAuthService,
        '_delOb',
        _delOb
    )
    wrap_method(
        PluggableAuthService,
        '_getAllLocalRoles',
        _getAllLocalRoles,
        add=True,
    )
    wrap_method(
        PluggableAuthService,
        '_doAddGroup',
        _doAddGroup,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        '_doAddUser',
        _doAddUser
    )
    wrap_method(
        PluggableAuthService,
        '_doChangeGroup',
        _doChangeGroup,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        '_doChangeUser',
        _doChangeUser,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        '_doDelGroups',
        _doDelGroups,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        '_doDelUser',
        _doDelUser,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        '_doDelUsers',
        _doDelUsers,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        '_getLocalRolesForDisplay',
        _getLocalRolesForDisplay,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        '_updateGroup',
        _updateGroup,
        add=True
    )
    wrap_method(
        PluggableAuthService,
        'addRole',
        addRole,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'authenticate',
        authenticate,
        add=True,
        roles=(),
    )
    wrap_method(
        PluggableAuthService,
        'canListAllGroups',
        canListAllGroups,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'canListAllUsers',
        canListAllUsers,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'credentialsChanged',
        credentialsChanged,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getAllLocalRoles',
        getAllLocalRoles,
        add=True,
    )
    wrap_method(
        PluggableAuthService,
        'getGroup',
        getGroup,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getGroupById',
        getGroupById,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getGroupByName',
        getGroupByName,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getGroupIds',
        getGroupIds,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getGroupNames',
        getGroupNames,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getGroups',
        getGroups,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getLocalRolesForDisplay',
        getLocalRolesForDisplay,
        add=True,
    )
    wrap_method(
        PluggableAuthService,
        'getUserIds',
        getUserIds,
        add=True,
        deprecated="Inefficient GRUF wrapper, use IUserIntrospection instead."
    )
    wrap_method(
        PluggableAuthService,
        'getUserNames',
        getUserNames,
        add=True,
        deprecated="Inefficient GRUF wrapper, use IUserIntrospection instead."
    )
    wrap_method(
        PluggableAuthService,
        'getUsers',
        getUsers,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'getPureUsers',
        getUsers,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'userFolderAddUser',
        postonly(userFolderAddUser),
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'userFolderDelUsers',
        postonly(_doDelUsers),
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'userFolderEditGroup',
        postonly(_doChangeGroup),
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'userFolderEditUser',
        postonly(_doChangeUser),
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'userFolderDelGroups',
        postonly(_doDelGroups),
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
    wrap_method(
        PluggableAuthService,
        'userSetGroups',
        _userSetGroups,
        add=True,
        deprecated="Method from GRUF was removed."
    )
    wrap_method(
        PluggableAuthService,
        'userSetPassword',
        userSetPassword,
        add=True,
        roles=PermissionRole(ManageUsers, ('Manager',))
    )
